/*

   lightgun.c

 */

#include <arch/arch.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>
#include <dc/maple/lightgun.h>
#include <string.h>
#include <assert.h>

CVSID("$Id: lightgun.c,v 1.5 2002/05/24 06:47:26 bardtx Exp $");

extern cont_btn_callback_t btn_callback;
extern uint8 btn_callback_addr;
extern uint32 btn_callback_btns;

unsigned char total_guns = 0;
int           gun_port[MAPLE_PORT_COUNT+1] = {-1,-1,-1,-1,-1};

static void lightgun_reply(maple_frame_t *frm) 
{
  maple_response_t  *resp;
  uint32            *respbuf;
  cont_cond_t       *raw;
  cont_state_t      *cooked;

  /* Unlock the frame now (it's ok, we're in an IRQ) */
  maple_frame_unlock(frm);

  /* Make sure we got a valid response */
  resp = (maple_response_t *)frm->recv_buf;

  respbuf = (uint32 *)resp->data;

  if (resp->response == MAPLE_RESPONSE_DATATRF)
  {
    /* -------------------------- */
    /* Standard button processing */
    /* -------------------------- */

    if (respbuf[0] != MAPLE_FUNC_CONTROLLER)
      return;

    /* Update the status area from the response */
    if (frm->dev) 
    {
      /* Verify the size of the frame and grab a pointer to it */
      assert( sizeof(cont_cond_t) == ((resp->data_len-1) * 4) );

      raw = (cont_cond_t *)(respbuf+1);

      /* Fill the "nice" struct from the raw data */
      cooked = (cont_state_t *)(frm->dev->status);
      cooked->buttons = (~raw->buttons) & 0xffff;
      cooked->ltrig = cooked->rtrig = 0;
      frm->dev->status_valid = 1;

      /* Check for magic button sequences */
      if (btn_callback) 
      {
        if (!btn_callback_addr ||
            (btn_callback_addr &&
             btn_callback_addr == maple_addr(frm->dev->port, frm->dev->unit))) 
        {
          if ( (cooked->buttons & btn_callback_btns) == btn_callback_btns ) 
          {
            btn_callback(maple_addr(frm->dev->port, frm->dev->unit),cooked->buttons);
          }
        }
      }
    }
  }
  else if (resp->response == MAPLE_RESPONSE_OK)
  {
    /* ----------------------- */
    /* Gun position processing */
    /* ----------------------- */

    if (frm->dev) 
    {
      int x,y;

      cooked = (cont_state_t *)(frm->dev->status);

      x = (*respbuf&0x3FF)-260;
      y = ((*respbuf>>15)&0x3FF)-40;

      if (x<0) x=0;
      else if (x>640) x=640;

      if (y<0) y=0;
      else if (y>480) y=480;

      /* -------------------------------------------------------------- */
      /* If the raw gun position is unchanged, set the x/y values to -1 */
      /* to indicate that the gun is off-screen.                        */
      /* -------------------------------------------------------------- */

      if ((x == cooked->joy2x) && (y == cooked->joy2y))
      {
        cooked->joyx = 0;
        cooked->joyy = 0;
      }
      else
      {
        cooked->joyx = (x/2)-160;
        cooked->joyy = (y/2)-120;

        if (cooked->joyx > 128) cooked->joyx = 128;
        else if (cooked->joyx < -128) cooked->joyx = -128;

        if (cooked->joyy > 128) cooked->joyy = 128;
        else if (cooked->joyy < -128) cooked->joyy = -128;
      }

      cooked->joy2x = x;
      cooked->joy2y = y;

      frm->dev->status_valid = 1;
    }
  }
}

static int lightgun_poll(maple_device_t *dev) {
	uint32 * send_buf;

	if (maple_frame_lock(&dev->frame) < 0)
		return 0;

	maple_frame_init(&dev->frame);
	send_buf = (uint32 *)dev->frame.recv_buf;
	send_buf[0] = MAPLE_FUNC_CONTROLLER;
	dev->frame.dst_port = dev->port;
	dev->frame.dst_unit = dev->unit;
	dev->frame.length = 1;
  dev->frame.callback = lightgun_reply;

  if (maple_state.dma_cntr%MAPLE_PORT_COUNT == dev->port)
	  dev->frame.cmd = 15;
  else
    dev->frame.cmd = MAPLE_COMMAND_GETCOND;

	dev->frame.send_buf = send_buf;
	maple_queue_frame(&dev->frame);

	return 0;
}

static void lightgun_periodic(maple_driver_t *drv) {
	maple_driver_foreach(drv, lightgun_poll);
}

static int lightgun_attach(maple_driver_t *drv, maple_device_t *dev) 
{
  /* ----------------------------------- */
  /* Try and cope with BIO gun wierdness */
  /* ----------------------------------- */

  if (dev->unit != 0)
    return -5;

	memset(dev->status, 0, sizeof(dev->status));

	dev->status_valid = 0;

  gun_port[total_guns++] = dev->port;

	return 0;
}

static void lightgun_detach(maple_driver_t *drv, maple_device_t *dev) 
{
  int i;

	memset(dev->status, 0, sizeof(dev->status));
	
  dev->status_valid = 0;

  /* ------------------------------- */
  /* shuffle the gun_port array down */
  /* ------------------------------- */

  for(i=0;i<total_guns;i++)
    if (gun_port[i] == dev->port)
      memcpy(&gun_port[i+1],&gun_port[i],((MAPLE_PORT_COUNT-i)*sizeof(int)));

  total_guns--;
}

/* Device Driver Struct */
static maple_driver_t lightgun_drv = {
	functions:	MAPLE_FUNC_LIGHTGUN,
	name:		"LightGun Driver",
	periodic:	lightgun_periodic,
	attach:		lightgun_attach,
	detach:		lightgun_detach
};

/* Add the lightgun to the driver chain */
int lightgun_init() {
	return maple_driver_reg(&lightgun_drv);
}

void lightgun_shutdown() {
	maple_driver_unreg(&lightgun_drv);
}

